﻿/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace MessageQueue
{
    class msmq
    {
        [StructLayout(LayoutKind.Sequential)]
        public struct MSGQUEUEOPTIONS
        {
            public uint dwSize;
            public uint dwFlags;
            public uint dwMaxMessages;
            public uint cbMaxMessage;
            public bool bReadAccess;
        };

        [DllImport("coredll.dll", EntryPoint = "WaitForSingleObject", SetLastError = true)]
        private static extern uint WaitForSingleObject(IntPtr hHandle, int wait);

        [DllImport("coredll.dll", EntryPoint = "CreateMsgQueue", SetLastError = true)]
        private static extern IntPtr CreateMsgQueue(string name, ref MSGQUEUEOPTIONS options);

        [DllImport("coredll.dll", EntryPoint = "ReadMsgQueue", SetLastError = true)]
        private static extern bool ReadMsgQueue(IntPtr hMsgQ, byte[] lpBuffer, uint cbBufSize, ref uint lpNumRead, int dwTimeout, ref uint pdwFlags);

        [DllImport("coredll.dll", EntryPoint = "WriteMsgQueue", SetLastError = true)]
        private static extern bool WriteMsgQueue(IntPtr hMsgQ, byte[] lpBuffer, uint cbBufSize, int dwTimeout, uint pdwFlags);

        [DllImport("coredll.dll", EntryPoint = "CloseMsgQueue", SetLastError = true)]
        private static extern bool CloseMsgQueue(IntPtr hMsgQ);

        public static uint MSGQUEUE_NOPRECOMMIT  = 0x00000001;
        public static uint MSGQUEUE_ALLOW_BROKEN = 0x00000002;

        static void Main(string[] args)
        {
            msmq q1 = new msmq(false);
            //System.Console.WriteLine("Queue1 created");
            msmq q2 = new msmq(false);
            //System.Console.WriteLine("Queue2 created");
            msmq q3 = new msmq(false);
            //System.Console.WriteLine("Queue3 created");
            msmq q4 = new msmq(false);
            //System.Console.WriteLine("Queue4 created");
            msmq q5 = new msmq(false);
            //System.Console.WriteLine("Queue5 created");
            msmq qr = new msmq(true);
            //System.Console.WriteLine("Receive Queue created");

            System.Threading.Thread threadr = new System.Threading.Thread(new System.Threading.ThreadStart(qr.listen));
            threadr.Start();
            //System.Console.WriteLine("Receive Queue listening");

            System.Threading.Thread thread1 = new System.Threading.Thread(new System.Threading.ThreadStart(q1.write));
            q1.setparams("queue1:", 30, 100);
            thread1.Start();

            System.Threading.Thread thread2 = new System.Threading.Thread(new System.Threading.ThreadStart(q2.write));
            q2.setparams("queue2:", 30, 110);
            thread2.Start();

            System.Threading.Thread thread3 = new System.Threading.Thread(new System.Threading.ThreadStart(q3.write));
            q3.setparams("queue3:", 30, 120);
            thread3.Start();

            System.Threading.Thread thread4 = new System.Threading.Thread(new System.Threading.ThreadStart(q4.write));
            q4.setparams("queue4:", 30, 90);
            thread4.Start();

            System.Threading.Thread thread5 = new System.Threading.Thread(new System.Threading.ThreadStart(q5.write));
            q5.setparams("queue5:", 30, 80);
            thread5.Start();

            System.Console.ReadLine();

            q1.close();
            q2.close();
            q3.close();
            q4.close();
            q5.close();

            qr.bQuit = true;
            System.Threading.Thread.Sleep(2000);
            qr.close();
        }

        private IntPtr hMSMQ = IntPtr.Zero;
        private bool bReadAccess = false;
        private bool bQuit = false;
        private int nTimes = 0;
        private string data = "";
        private int nSleep = 0;
        public msmq(bool bRead)
        {
            bReadAccess = bRead;
            MSGQUEUEOPTIONS options = new MSGQUEUEOPTIONS();
            options.dwFlags = MSGQUEUE_NOPRECOMMIT | MSGQUEUE_ALLOW_BROKEN;
            options.dwMaxMessages = 0;
            options.cbMaxMessage = 1024;
            options.bReadAccess = bReadAccess;
            options.dwSize = (uint)System.Runtime.InteropServices.Marshal.SizeOf(options);
            hMSMQ = CreateMsgQueue("Test", ref options);
        }

        public void setparams (string _data, int _nTimes, int _nSleep)
        {
            data = _data;
            nTimes = _nTimes;
            nSleep = _nSleep;
        }

        public void write()
        {
            while (nTimes-- > 0)
            {
                string s = data + nTimes.ToString();
                byte[] bdata = System.Text.Encoding.ASCII.GetBytes(s);
                uint uSize = (uint)bdata.Length;
                WriteMsgQueue(hMSMQ, bdata, uSize, 500, 0);
                System.Threading.Thread.Sleep(nSleep);
            }
            System.Console.WriteLine(data + "closed");
        }

        public void listen()
        {
            byte[] buf = new byte[1024];
            uint nRead = 0, flags = 0;

            if (bReadAccess)
            {
                while (!bQuit)
                {
                    if (WaitForSingleObject(hMSMQ, 1000) == 0)
                    {
                        flags = 0;
                        ReadMsgQueue(hMSMQ, buf, (uint)buf.Length, ref nRead, -1, ref flags);
                        string msg = System.Text.Encoding.ASCII.GetString(buf, 0, (int)nRead);
                        //System.Console.WriteLine("R:" + msg + " flag=" + flags.ToString());
                        System.Console.WriteLine("Received:" + msg);
                    }
                }
            }
            System.Console.WriteLine("Receive Queue closed");
        }

        public void close()
        {
            CloseMsgQueue(hMSMQ);
        }
    }
}
